// Copyright (C) 1999 Fredrik Ohrstrom foh@pureinteraction.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Library General Public
//    License as published by the Free Software Foundation; either
//    version 2 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Library General Public License for more details.
//
//    You should have received a copy of the GNU Library General Public
//    License along with this library; if not, write to the Free
//    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
//    02111-1307, USA

#ifndef _Component_Implementations_hh
#define _Component_Implementations_hh

#include <semaphore.h>
#include <list>
#include <map>
#include <multimap.h>
#include "string.hh"
#include "buffer.hh"
#include "Component_Interfaces.hh"

namespace Component_Implementations
{
        using namespace Component_Interfaces;
        
        extern Resources *resources;
        extern char *resource_filename;
        extern void useResourceFile (char *fn);

        class MetaClass 
        {};

        class ClassType
        {
                public:
                
                string type_;

                ClassType (string type);
        };
        
        class Component_i : public virtual Component,
                            public virtual ClassType
        {
                int magic;                
                list<Component_ptr> components_;
                Component_ptr composite_;
                string name_;
                string path_;
                string nameandtype_;
                Factory_ptr factory_;
                
                public:
                
                Component_i () : ClassType (Latin1("Component_i")) {}
                
                Component_i (Component_ptr composite,
                             string name);
                virtual ~Component_i ();

                Component_ptr cleanup ();
                void _dispose ();
                Component *composite ();
                list<Component_ptr> components ();
                string path ();
                string nameAndType ();
                string name ();                
                void addChild (Component_ptr c);
                void delChild (Component_ptr c);
                void updateResources ();
                Factory_ptr factory ();
                void printTree ();
        };
        
        struct SysCaller 
        {
                virtual void callit (Component_i *o,
                                     Component_i *c,
                                     Component_i *d) = 0;
        };
        
        struct Flowback_i : public virtual Flowback,
                            public Component_i
        {
                SysCaller *sys_caller_;
                MetaClass *data_type_;
                Component_ptr invoker_;
                
                Flowback_i (Component *c,
                            SysCaller *sc,
                            MetaClass *type)
                                : Component_i (c,Latin1("")),
                                  ClassType (Latin1("Flowback_i")),
                                  sys_caller_ (sc),
                                  data_type_(type)
                {
                }
                
                void invokeflow (Component *o)
                {
                        if (sys_caller_)
                                sys_caller_->callit (
                                        dynamic_cast<Component_i*>(composite()),
                                        this,
                                        dynamic_cast<Component_i*>(o));
                }
                
                int  handle (Component *o)
                {
                        return 1;
                }

                void setInvoker (Component_ptr c)
                {
                        invoker_ = c;
                }
                
                Component_ptr getInvoker ()
                {
                        return invoker_;
                }
        };
        
        struct Callback_i : public virtual Callback,
                            public Flowback_i
        {
                Callback_i (Component_i *c,
                            SysCaller *sc)
                                : Flowback_i (c,sc,0),
                                  ClassType (Latin1("Callback_i"))
                {
                }
                
                void invoke ()
                {
                        if (sys_caller_)
                        {
                                sys_caller_->callit (
                                        dynamic_cast<Component_i*>(composite()),
                                        this,
                                        0);
                        }
                }                
        };
                
        template<class T, class C, class D>
        struct Flower : public SysCaller
        {
                typedef void (T::*Function)(C*,D*);
	     
                Flower (Function function)
                   : function_ (function)
                { }
	     
                void callit (Component_i *o,
                             Component_i *c,
                             Component_i *d)
                { 
                        (((T*)o)->*function_) ((C*)c,(D*)d);
                }

                bool isEqualTo (SysCaller *c)
                {
                        return (function_==((Flower*)c)->function_);
                }

                Function function_;
        };
        
        template<class T, class C, class D>
        inline 
        C* flow (T *object, void (T::*function)(C*,D*))
        {
                return new C (object, 
			 new Flower<T,C,D> (function),
			 0);   
        }

        template<class T, class C>
        struct Caller : public SysCaller
        {
                typedef void (T::*Function)(C*);
                
                Caller (Function function) : function_ (function) { }
                
                void callit (Component_i *o,
                             Component_i *c,
                             Component_i *d)
                {
                        (((T*)o)->*function_) ((C*)c);
                }
                bool isEqualTo (SysCaller *c)
                {
                        return (function_==((Caller*)c)->function_);
                }
                
                Function function_;
        };

        template<class T, class C>
        inline 
        C *call (T *object, void (T::*function)(C*))
        {
                return new C (object, new Caller<T,C> (function));
        }

        class Mutex 
        {
                public:

                Mutex (string name=Latin1(""));
                void lock (char *p=0);
                void unlock (char *p=0);
                int trylock (char *p=0);
                
                private:

                pthread_mutex_t mutex_;
                string name_;
        };

        class Lock
        {
                public:

                Lock (Mutex *m)
                                : mutex_ (m)
                {
                        mutex_->lock ();                        
                }
                ~Lock ()
                {
                        mutex_->unlock ();
                }
                Mutex *mutex () 
                {
                        return mutex_;
                }

                private:
                
                Mutex *mutex_;
        };
                
        class RecursiveMutex 
        {
                public:

                RecursiveMutex (string name);
                void lock (char *p=0);
                void unlock (char *p=0);
                
                private:

                Mutex structure_lock_;
                Mutex lock_;
                pthread_t owner_;
                int nof_locks_;
                string name_;
        };

        class Semaphore
        {
                public:

                Semaphore (string name=Latin1(""));
                ~Semaphore ();
                void post (char *p=0);
                void wait (char *p=0);
                int  value ();
                
                private:

                sem_t semaphore_;
                string name_;
        };

        class Thread : public Component_i
        {
                public:

                Thread (Component_i *p=0, string name=Latin1(""));
                ~Thread ();
                void start (Callback *cb);
                void restart ();
                void stop ();
                bool running () { return running_; }
                static int  nofThreads ();
                static Thread *getThread (int n);	     
                static void setMainThread ();
                
                private:

                pthread_t     thread_;
                int           pid_;
                bool          running_;
                Callback_ptr  called_;
                
                static void *start_thread (Thread *thread);
                
                static Thread *all_threads_[65536];
                static int nof_threads_;
                static Mutex protect_thread_info_;
        };
        
        struct DataReceiver
        {
                virtual void receive (buffer b) = 0;
        };        
        
        class SelectController : public Component_i
        {
                fd_set read_fds_;
                fd_set write_fds_;
                fd_set exception_fds_;
                
                DataReceiver *read_fd_ [64];
                DataReceiver *write_fd_ [64];
                DataReceiver *exception_fd_ [64];
                
                int fildes[2];
                
                int my_pid;
                Thread thread_;
                
                void run (Callback_i *cb);

                public:

                SelectController ();
                ~SelectController ();
                void addConnection (DataReceiver *dr, int n);
                void shake ();
        };

        struct ResourceRow
        {
                ResourceRow (string m=Latin1(""), string me=Latin1(""),
                             string mm=Latin1(""), string d=Latin1(""))
                                : match(m), matchend(me), matchmiddle (mm), data(d) {}                

                string match, matchend, matchmiddle;
                string data;
        };

        struct Resources_i : public virtual Resources,
                             public Component_i
        {
                multimap<string,ResourceRow*> table;
                map<string,ResourceRow*> exact_table;
                
                Resources_i (Component *c,
                             string name,
                             string filename)
                                : Component_i (c, name),
                                  ClassType (Latin1("Resources"))
                {
                        char fn [256];
                        UniToUtf (filename, fn);                        
                        loadFile (fn);
                        fprintf (stderr, "Loaded resource file.\n");
                }

                multimap<int,ResourceRow*> findResources (string match);                                
                ResourceRow *findResource (string match);
                
                string whichResource (string match);
                string getResource (string match);
                string getExactResource (string match);                
                list<string> getResources (string match);

                void setResource (string match,
                                  string data);

                void loadFile (const char *name);
                void saveToFile (char *name);
        };

        struct Integer : public Component_i
        {
                int value_;

                public:                

                Integer (Component * c,
                         string n,
                         int v);

                int value ();
        };

        struct String : public Component_i
        {
                string value_;

                public:                

                String (Component * c,
                        string n,
                        string v);
                
                string value ();
        };

        struct StringMap : public Component_i
        {
                map<string,string> value_;

                public:

                StringMap (Component_ptr c,
                           string n);
                
                map<string,string> &value ();
        };
        
        struct Factory_i : public virtual Factory,
                           public Component_i
        {
                void disposeComponent (Component_ptr component);

                Factory_i (Component_ptr composite, string name);
                
                /*                
                Component_ptr newRootComponent ();                
                Component_ptr newComponent (Component_ptr container);
                Component_ptr newComponentArg (Component_ptr container,
                Component_ptr arg); */
        };        
}

#endif







