// Copyright (C) 1997 Fredrik Ohrstrom /////////////////////////////////////
// Copyright (C) 1999 Fredrik Ohrstrom /////////////////////////////////////
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Library General Public
//    License as published by the Free Software Foundation; either
//    version 2 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Library General Public License for more details.
//
//    You should have received a copy of the GNU Library General Public
//    License along with this library; if not, write to the Free
//    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
//    02111-1307, USA
//
// File     : array.H
// Contents : An array class.
//            These objects are not distributable.
// Classes  : array<> array<>::iter

#ifndef _array_H
#define _array_H

#include "ref.H"

template<class T>
class static_array
{
    public:  static_array (T *objects, int n) 
                : objects_ (objects), nof_objects_(n)
	     {
	     }

	     T &operator [] (int i)
	     {
		 return objects_[i];
	     }

	     int size ()
	     {
		 return nof_objects_;
	     }
	
    private: T   *objects_;
	     int  nof_objects_;
};

template<class T>
class array
{
  public:

  array (int buffer_size = 256);

  void append  (T t);
  void append  (array a);
  void clear ();
  T   *begin () { return data_->begin_; }
  T   *end ()   { return data_->end_-1; }

  array operator + (array a);
  bool  inArray (int i) { return (i>=0 && i<data_->size_); }
  T    &operator [] (int i) { if (i<0 || i>=data_->size_) return (T(0));
                              return data_->begin_[i]; }
  array copy ();
  bool  isEmpty ();
  int   size ();
  void  setSize(int new_size);

  struct arrayData;

  class iter
  {
    public:

    iter () : node_ (0), array_ (0) { }

    void useArray    (array &array) {array_=array.data_;node_=array_->begin_;} 
    void rewind      () { if (array_) node_ = array_->begin_; }
    void rewend      () { if (array_) node_ = array_->end_-1; }
    bool inArray     () { return (node_>=array_->begin_ && 
				  node_<array_->end_)?true:false; }
    void operator ++ () { node_++; }
    void operator -- () { node_--; }
    T   &operator *  () { return *node_; }
    T   &curr ()        { return *node_; }
    T   &next ()        { return *(node_+1); }
    T   &prev ()        { return *(node_-1); }
    bool operator == (iter &i) { return node_==i.node_ && array_==i.array_; }
    bool operator != (iter &i) { return node_!=i.node_ || array_!=i.array_; }
    void go (int i) { node_ = array_->begin_+i; }
    int  pos () { return node_-array_->begin_; }
 
    void useArrayData (arrayData *data) { array_=data;node_=array_->begin_; }

    private:

    T *node_;
    arrayData *array_;
  };

  private:

  struct arrayData : public ref
  {
    arrayData (int buffer_size) 
      : begin_ (new T [buffer_size]), 
	buffer_size_ (buffer_size),
        size_ (0) 
    { 
      end_ = begin_;
    }
    ~arrayData ()
    {
      delete [] begin_;
    }
    void ResizeBuffer (int new_size);

    T *begin_, *end_;
    int size_, buffer_size_;
  };

  lptr<arrayData> data_;
};

template<class T> inline
void array<T>::arrayData::ResizeBuffer (int new_size)
{
  T *tmp = new T [new_size];
  if (new_size<size_) size_ = new_size;
  memcpy ((char*)tmp, (char*)begin_, size_*sizeof(T));
  delete [] begin_;
  buffer_size_ = new_size;
  begin_ = tmp;
  end_ = tmp+size_;
}

template<class T> inline
array<T>::array (int buffer_size) 
  : data_ (new arrayData (buffer_size)) { }

template<class T> inline
void array<T>::append (T t)
{
  if (data_->size_ >= data_->buffer_size_)
  {
    data_->ResizeBuffer (data_->buffer_size_*2);
  }

  *data_->end_ = t;
  data_->end_++;
  data_->size_++;
}  

template<class T> inline
void array<T>::append (array a)
{
  iter i;
  for (i.useArray (a); i.inArray (); ++i) append (*i);
}  

template<class T> inline
void array<T>::setSize (int new_size)
{
  data_->ResizeBuffer (new_size);
  data_->size_ = new_size;
  data_->end_ = data_->begin_+data_->size_;
}

template<class T> inline
void array<T>::clear ()
{
  array empty (0);
  *this = empty;
}

template<class T> inline
array<T> array<T>::copy ()
{
  array l;
  l.setSize (data_->size_);
  memcpy (l.data_->begin_, data_->begin_, data_->size_*sizeof(T));
  return l;
}

template<class T> inline
array<T> array<T>::operator + (array a)
{
  array b;
  b.append (*this);
  b.append (a);
  return b;
}

template<class T> inline
bool array<T>::isEmpty ()
{
  return !data_->size_;
}

template<class T> inline
int array<T>::size ()
{
  return data_->size_;
}

#endif
