// Copyright (C) 1999 Fredrik Ohrstrom
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Library General Public
//    License as published by the Free Software Foundation; either
//    version 2 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Library General Public License for more details.
//
//    You should have received a copy of the GNU Library General Public
//    License along with this library; if not, write to the Free
//    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
//    02111-1307, USA

#include"tcp.hh"

using namespace Tcp;


// Function: waitForReadData 
//  Purpose: Sleeps until there is data available for reading
//           on the supplied file descriptor.
int waitForReadData (int fd)
{
        int n;                
        ioctl (fd, FIONREAD, &n);
        if (n>0) return 1;

        fd_set rfds, wfds, efds;
        
        for (int i=0; i<64; ++i)
        {
                FD_CLR (i,&rfds);
                FD_CLR (i,&wfds);
                FD_CLR (i,&efds);
        }
        
        FD_SET (fd,&rfds);
        
        select (64, &rfds, &wfds, &efds, NULL);
        
        if (FD_ISSET (fd,&rfds))
        {
                return 1;
        }
        return 0;
}

//  Method: TCPServer
// Purpose: Creates a server object that monitors a port.
Tcp::TCPServer::TCPServer (int port)
                : Component_i (0,Latin1("Urk")), ClassType (Latin1("TCPServer")),
                  accept_thread (this, Latin1("accept_thread"))
{
        signal (SIGPIPE, SIG_IGN);
        
        sockfd = socket (AF_INET, SOCK_STREAM, 0);
        
        if (sockfd<0)
        {
                fprintf (stderr, "Tcpserver can't open stream socket.\n");
                exit (-1);
        }
        
        memset ((char *) &serv_addr, 0, sizeof(serv_addr));
        serv_addr.sin_family = AF_INET;
        serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        serv_addr.sin_port = htons (port);
        
        int foo=1;
        setsockopt (sockfd, SOL_SOCKET, SO_REUSEADDR,
                    (const char*)&foo, sizeof(foo));

        int r = bind(sockfd, (struct sockaddr *) &serv_addr, 
                     sizeof(serv_addr));
        
        if (r<0)
        {
                fprintf (stderr, "Tcpserver can't bind local address.\n");
                exit (-1);
        }
        
        listen (sockfd, 10);        
        accept_thread.start (call (this, &Tcp::TCPServer::acceptConnections));


        /*
          To be done i desctructor
            SSL_CTX_free (ctx);
        */
}

void Tcp::TCPServer::acceptConnections (Callback_i *cb)
{	     
        for (int i=0; i<10; ++i)
        {
                freelist.push_back (new Client (this));
                waiting_for_clients.post ();
        }

        fprintf (stderr, "Accepting connections...%d threads\n",
                 waiting_for_clients.value());
        
	while (accept_thread.running())
	{
                clilen = sizeof(cli_addr);
                newsockfd = accept (sockfd,
                                    (struct sockaddr*)&cli_addr,
                                    (unsigned*)&clilen);
                
                if (newsockfd<0)
                {
                        cerr << "Server: accept error" << endl;
                        exit (-1);
                }

                dispatchClient (newsockfd);
	}
}

void Tcp::TCPServer::dispatchClient (int fd)
{
        //        fprintf (stderr, "Threads available %d\n", waiting_for_clients.value());
        waiting_for_clients.wait ();
        freelist_protect.lock ();
        Client *c = freelist.back ();
        freelist.pop_back ();        
        freelist_protect.unlock ();
        c->start (fd);
        
}

void Tcp::TCPServer::returnClient (Client *c)
{
        freelist_protect.lock();
        freelist.push_back (c);
        freelist_protect.unlock();
        waiting_for_clients.post ();
}
                
Tcp::Client::Client (TCPServer *server)
                : Component_i (0, Latin1("Client")), ClassType (Latin1("Client"))
{
        server_ = server;
        fd_ = -1;
        echo_thread_.start (call (this, &Tcp::Client::serveClient));
}

void Tcp::Client::start (int fd)
{
        fd_ = fd;
        nap_.post ();
}

void Tcp::Client::stop ()
{
        if (fd_>=0) close (fd_);
        fd_=-1; 
        
        server_->returnClient (this);
}

void Tcp::Client::serveClient (Callback_i *cb)
{
        int rr, n;

        while (echo_thread_.running ())
        {
                // Sleep until needed
                nap_.wait ();
                
                /*                rr = waitForReadData (fd_);
                if (rr)
                { */
                        /*                        ssl = SSL_new (server_->ctx);
                        CHK_NULL(ssl);
                        SSL_set_fd (ssl, fd_);
                        int err = SSL_accept (ssl);
                        CHK_SSL(err);
                        fprintf (stderr, "SSL connection using %s\n", SSL_get_cipher (ssl));
                        */
                        server_->handler()->handle (this);
                        /*                } */
                stop ();
        }        
}

int hextoint (unichar a)
{
        if (a>='0' && a<='9') return a-'0';
        if (a>='A' && a<='F') return a-'A'+10;
        if (a>='a' && a<='f') return a-'a'+10;
        return 0;
}

string translatechars (string v)
{
        string r;
        
        int l = v.size();
        for (int i=0; i<l; ++i)
        {
                switch (v[i])
                {
                    case '%':
                            if (i<(l-2))
                            {                                    
                                    unichar k = hextoint(v[i+1])*16+hextoint(v[i+2]);
                                    r = r+k;
                                    i+=2;
                            }
                            break;
                    case '+':
                            r = r+Latin1(" ");
                            break;
                    default:
                            r = r+v[i];
                }
        }
        return r;
}

string Tcp::TranslateHtmlSensitiveChars (string v)
{
        string r;
        
        int l = v.size();
        for (int i=0; i<l; ++i)
        {
                switch (v[i])
                {
                    case '&':
                            r = r+Latin1("&amp;");
                            break;
                    case '<':
                            r = r+Latin1("&lt;");
                            break;
                    case '>':
                            r = r+Latin1("&gt;");                            
                            break;
                    case '"':
                            r = r+Latin1("&quot;");
                            break;
                    case '%':
                            r = r+Latin1("&#37;");
                            break;
                    default:
                            r = r+v[i];
                }
        }
        return r;
}

unichar hexToDec (unichar a)
{
        if (a>='a' && a<='f') return a-'a'+10;
        if (a>='A' && a<='F') return a-'A'+10;        
        if (a>='0' && a<='9') return a-'0';
        return 0;
}

string convertHex (unichar a, unichar b)
{
        a = hexToDec(a);
        b = hexToDec(b);
        unichar t = (unichar)(a*16+b);
        return Latin1("")+t;
}

string Tcp::TranslateHtmlSensitiveCharsBack (string v)
{
        string r;
        
        int l = v.size();
        for (int i=0; i<l; ++i)
        {
                switch (v[i])
                {
                    case '%':
                            r = r+convertHex (v[i+1],v[i+2]);
                            i+=2;
                            break;
                    default:
                            r = r+v[i];
                }
        }
        return r;
}

void parseRequest (char *i, map<string,string> &values)
{
        char msg[32000];
        //        fprintf (stderr, "Request start-------------\n%s\nRequest end------------------\n", i);
        strncpy (msg, i, 32000);        
        string in = Latin1(msg);
        
        int eoll = in.findSubstring (0, Latin1("\n"));
        string getrow = in.substring (0,eoll);
        int hp = getrow.findSubstring (0,Latin1(" HTTP/1.0"));
        if (hp>1) getrow = getrow.substring (0, hp);
        getrow = getrow.trimends ();

        getrow.print (stderr, "MSG>", "<\n");
        
        int p = getrow.findSubstring (0, Latin1("/"));
        int pp = getrow.findSubstring (p+1, Latin1("/"));
        int ppp = getrow.findSubstring (pp+1, Latin1("?"));


        fprintf (stderr, "p=%d pp=%d ppp=%d\n", p,pp,ppp);        
        if (ppp<0) ppp = getrow.size();
        
        if (pp<0 && p>0)
                values[Latin1("cookie")] = getrow.substring (p+1);
        if (pp>p && p>0)
                values[Latin1("cookie")] = getrow.substring (p+1,pp);
        if (ppp>pp && pp>p && p>0)
                values[Latin1("state")] = getrow.substring (pp+1,ppp);

        if (getrow.size()>ppp && ppp>pp && pp>p && p>0)
        {
                string messages = getrow.substring (ppp+1,0);
                
                int eol = messages.findSubstring (0, Latin1(" "));
                if (eol<0) eol=messages.size();
                int blockstart=0, epos=0, blockend=0;
                for (;;)
                {
                        epos = messages.findSubstring (blockstart,Latin1("="));
                        if (epos<0) break;
                        blockend = messages.findSubstring (epos, Latin1("&"));

                        if (blockend<0 || blockend>=eol) blockend=eol;
                        
                        if (blockstart<=(epos+1) && (epos+1)<blockend)
                        {
                                string val = messages.substring (epos+1, blockend);
                                val = translatechars (val);
                                values [messages.substring (blockstart,epos)]
                                = val;
                        }
                        blockstart=blockend+1;
                        if (blockstart>=messages.size()) break;
                }
        }
}

        




















